/*     Utilidade para ver pantallas de Spectrum e convertilas de/a GIF     */
/*                  Programada por Alberto Garcia Gonzalez                 */
/*         Modulo para traballar con ficheiros .GIF 256x192x256            */

/* ULTIMA REVISION: 3-3-99 21:55 */

#define ROOT -1
#define CC (pot2[bpp])
#define EOI (pot2[bpp]+1)

struct entrada {        /* Cada unha das entradas do diccionario */
      int cabec;
      byte cola;
};

struct codigo {
      byte datos[4096];
      int inicio;
};

void descomprimir(int dato);
int posicion_tabla(struct entrada dato,int max_root,int max_tabla);
void agregar_dato(dword dato,byte *buffer,dword *bit_actual,byte nbits);
bytepunt quitar_entrelaz(bytepunt buffer,int x,int y);

/* Estas son variables globales para evitar pasalas como parametro */

int pot2[13] = { 1,2,4,8,16,32,64,128,256,512,1024,2048,4096 };
struct entrada tabla[4096];                              /* O diccionario */
struct codigo code;                 /* O dato unha vez esta descomprimido */

/***************************************************************************/

bytepunt leer_gif(char nome[],int *x, int *y)
{
 FILE *fich;            /* O ficheiro do que lemos */
 bytepunt buffer;       /* A pantalla descomprimida */
 bytepunt datosgif;     /* O ficheiro a descomprimir */
 byte temporal[255];    /* Buffer temporal onde leremo-la cabeceira */
 int i,datotabla;       /* O dato libre da tabla */
 byte nbits;            /* Os datos que vamos lendo do ficheiro */
 byte bpp;              /* Numero de bits de profundidade */
 dword pos;             /* A posicion que escribimos na pantalla */
 dword tam_datos;       /* O tamanho de todolos datos comprimidos */
 byte tam_bloque;       /* O tamanho de cada bloque de datos */
 byte entrelaz;         /* Indica se a imaxe esta entrelazada */
 int dato;              /* O dato que queremos descomprimir */
 int old;               /* Codigo anterior */
 dword tamanho;
 dword bit_actual;      /* O bit dos datos onde debemos ler */
 dword grande;          /* 32 bits para extrae-lo dato */

 if ((fich = fopen(nome,"rb")) == NULL) {
       error1(err_apertura,nome);
 }
 tamanho=filelength(fileno(fich));

 leer(temporal,13,fich);    /* Cabeceira e Logical Screen Descriptor Block */

 if ((temporal[0] != 'G') || (temporal[1] != 'I') || (temporal[2] != 'F'))
       error(non_e_GIF);

 bpp = (temporal[10]&7) + 1;           /* Lemo-los bits por pixel da imaxe */

 leer(paleta,3*pot2[bpp],fich);      /* Paleta VGA */
 for (i=0;i<3*pot2[bpp];i++) paleta[i] >>= 2;
 leer(temporal,1,fich);

 tamanho -= 3*pot2[bpp];

 if (temporal[0] == 0x21) {     /* Image Extension Block */
       leer(temporal,1,fich);
       tamanho--;
       do {
             leer(temporal,1,fich);
             tamanho--;
             if (temporal[0]) {
                   tamanho-=temporal[0];
                   leer(temporal,temporal[0],fich);
             }
             else break;
       } while (1);
       leer(temporal,1,fich);
 }

 leer(temporal+1,10,fich);    /* Image Descriptor Block */

 if ((temporal[0] != 0x2c) || ((*x=temporal[5]+256*temporal[6]) > 256) ||
      ((*y=temporal[7]+256*temporal[8]) > 192))
    error(tam_img_inc);
   
 if (temporal[9] & 128) error(non_soport);   /* Non admite Local Color Map */
 
 entrelaz = (temporal[9] & 64);

 if ((temporal[10] == 2) && (bpp == 1)) {
    bpp++;           /* Caracteristica especial dos GIFs en branco e negro */
 }

 tamanho-=25;
 datosgif = (bytepunt) malloc(tamanho);
 buffer = (bytepunt) malloc (64000);
 
 if ((buffer == NULL) || (datosgif == NULL)) error(non_hai_mem);

 leer(&tam_bloque,1,fich);
 tam_datos=0;                                /* Todolos bloques a memoria */
 while (tam_bloque) {
       leer(temporal,tam_bloque,fich);
       bpmemcpy(datosgif+tam_datos,temporal,tam_bloque);
       tam_datos += tam_bloque;
//       for (i=0;i<tam_bloque;) datosgif[tam_datos++]=temporal[i++];
       leer(&tam_bloque,1,fich);
 }
 fclose(fich);

 for (i=0;i<pot2[bpp];i++) {        /* Inicializamo-las variables */
       tabla[i].cabec = ROOT;
       tabla[i].cola = i;
 }
 datotabla=EOI+1; nbits=bpp+1;
 bit_actual=0; pos=0;
 old = -1;

                      /* O cotarro empeza aqui */

 do {
       grande=65536*(dword)datosgif[2+(bit_actual>>3)]+
              256*(dword)datosgif[1+(bit_actual>>3)]+
              (dword)datosgif[bit_actual>>3];
       grande >>= (bit_actual&7);
       dato=grande&(pot2[nbits]-1);
       bit_actual+=nbits;
       if (dato == CC) {
             nbits=bpp+1;
             datotabla=EOI+1;
             old=-1;
       }
       else if (dato != EOI) {
             if (dato<datotabla) {         /* Se o dato esta na tabla */
                   descomprimir(dato);
                   for (i=code.inicio;i<4096;) buffer[pos++]=code.datos[i++];
                   if (old != -1) {
                         tabla[datotabla].cabec=old;
                         tabla[datotabla].cola=code.datos[code.inicio];
                         if (++datotabla==pot2[nbits]) nbits++;
                   }
             }
             else {                   /* Se o dato non esta na tabla */
                   if (old != -1) {
                         descomprimir(old);
                   }
                   else {
                         error(err_datos);
                   }
                   tabla[datotabla].cabec=old;
                   tabla[datotabla].cola=code.datos[code.inicio];
                   if (++datotabla==pot2[nbits]) nbits++;
                   for (i=code.inicio;i<4096;)
                         buffer[pos++]=code.datos[i++];
                   buffer[pos++]=code.datos[code.inicio];
             }
             old = dato;
             if (nbits>12) {   /* Suponse que esto non poderia pasar */
                   nbits=12;   /* Pero se non se pon, casca :-? */
             }
       }          /* else */
 } while ((dato != EOI) && (pos<64000));
 while (pos<((dword)(*x)*(*y))) buffer[pos++]=code.datos[code.inicio];
 free(datosgif);
 
 if (entrelaz) buffer = quitar_entrelaz(buffer,(*x),(*y));
 
 return buffer;
}

/***************************************************************************/

void descomprimir(int dato)
{
 int pos=4095;
 do {
       code.datos[pos--]=tabla[dato].cola;
       dato = tabla[dato].cabec;
 } while (dato != ROOT);
 code.inicio=pos+1;
}

/***************************************************************************/

bytepunt comprimir_gif(bytepunt orig,dword *tamanho)
{
 byte buffer[25000];                     /* Buffer temporal */
 bytepunt final;                         /* Resultado final */
 char cabec[]="GIF87a\0\x1\xc0\0\xb3\0\0";
 char idb[]=",\0\0\0\0\0\x1\xc0\0\x3\x4";
 dword tam;
 dword bit_actual;
 dword grande;                       /* Onde se fan calculos temporalmente */
 int datolibre;	                          /* Numero do dato libre na tabla */
 struct entrada antiguo;                                  /* Dato anterior */
 byte bpp=4;       /* 4 bpp son suficientes para unha pantalla de Spectrum */
 byte nbits;
 int i;
 unsigned int pos;        /* Posicion do buffer orixinal que estamos lendo */
 int nbloques;
 int pos_dato_actual;      /* Posicion na tabla do dato que imos comprimir */
 
 memcpy(buffer,cabec,13);
// for (i=0;i<13;i++) buffer[i]=(byte) cabec[i];
 for (i=0;i<16;i++) {
       buffer[13+3*i]=colores[i][0]<<2;
       buffer[13+3*i+1]=colores[i][1]<<2;
       buffer[13+3*i+2]=colores[i][2]<<2;
 }
 memcpy(buffer+61,idb,11);
// for (i=0;i<11;i++) buffer[i+61]=(byte) idb[i];
 memset(buffer+72,0,25000-72);          /* Vaciamos o buffer */

 for (i=0;i<pot2[bpp];i++) {        /* Inicializamos a tabla */
       tabla[i].cabec = ROOT;
       tabla[i].cola = i;
 }
 datolibre = EOI+1;
 nbits = bpp+1;
 bit_actual = 0;
 pos = 0;
 antiguo.cabec = ROOT;
 agregar_dato(CC,buffer,&bit_actual,nbits);

 /* Aqui comeza a compresion */

 while (pos<49152) {                              /* O tamanho da pantalla */
       antiguo.cola = orig[pos++];
       if ((pos_dato_actual=posicion_tabla(antiguo,EOI+1,datolibre-1))!=-1) {
             antiguo.cabec = pos_dato_actual;
       }
       else {
             tabla[datolibre++] = antiguo;
             grande = antiguo.cabec;
             antiguo.cabec = antiguo.cola;
             agregar_dato(grande,buffer,&bit_actual,nbits);
             if (datolibre == pot2[nbits] + 1) {
                   nbits++;
             }
             else if (datolibre == 4096) {
                   nbits = bpp+1;
                   datolibre = EOI+1;
//                   antiguo.cabec = ROOT;  NOOOOOONNN!!!
                   agregar_dato(CC,buffer,&bit_actual,12);
             }
       }
 }
 if (antiguo.cabec != ROOT) {
       agregar_dato(antiguo.cabec,buffer,&bit_actual,nbits);
 }
 agregar_dato(EOI,buffer,&bit_actual,nbits);
 
 tam = bit_actual >> 3;
 if (bit_actual & 7) tam++;
 i = (tam%255) ? 1 : 0;
 nbloques = (tam/255);                   /* Numero de bloques de 255 bytes */
 *tamanho = tam + nbloques + 73 + i;
 
 if ((final = (bytepunt) malloc (*tamanho)) == NULL) {
       error(non_hai_mem);
 }

 bpmemcpy(final,buffer,72);
// for (grande=0;grande<72;grande++) final[grande] = buffer[grande];
 for (grande=0;grande<nbloques;grande++) {
       final[72+grande*256] = 255;
       bpmemcpy(final+73+grande*256,buffer+72+grande*255,255);
/*       for (i=0;i<255;i++) {
             final[73+grande*256+i] = buffer[72+grande*255+i];
       } */
 }
 grande = nbloques;
 nbloques = tam%255;                  /* Tamanho do ultimo bloque de bytes */
 if (nbloques) {
       final[72+grande*256] = nbloques;
       bpmemcpy(final+73+grande*256,buffer+72+grande*255,nbloques);
/*       for (i=0;i<nbloques;i++) {
             final[73+grande*256+i] = buffer[72+grande*255+i];
       } */
 }
 final[*tamanho-1] = 0;
 return final;
}

/***************************************************************************/

void agregar_dato(dword dato,byte *buffer,dword *bit_actual,byte nbits)
{
 dword pos;
 byte posbit;
 dword bit;

 bit = *bit_actual;
 pos = bit >> 3;
 posbit = bit & 7;
 buffer[72+pos] |= (255 & (dato << posbit));
 buffer[73+pos] |= (255 & (dato >> (8-posbit)));
 buffer[74+pos] |= (255 & (dato >> (16-posbit)));
 *bit_actual += nbits;
}

/***************************************************************************/

int posicion_tabla(struct entrada dato,int max_root,int max_tabla)
{
 int i;
 i = max_root<dato.cabec ? dato.cabec+1 : max_root;
 if (dato.cabec == ROOT)
       return dato.cola;
 for (;i <= max_tabla;i++) {
       if ((tabla[i].cabec == dato.cabec) && (tabla[i].cola == dato.cola))
             return i;
 }
 return -1;
}

/***************************************************************************/

bytepunt quitar_entrelaz(bytepunt buffer,int x,int y)
{
 bytepunt temporal;
 byte barrido;                         /* O barrido que estamos procesando */
 byte inicio[] = { 0, 4, 2, 1};         /* Linha de inicio de cada barrido */
 byte salto[] = { 8, 8, 4, 2};  /* Numero de linhas que se saltan cada vez */
 dword pos_orig = 0;                        /* Posicion no buffer de orixe */
 dword pos_dest;                          /* Posicion no buffer de destino */
 int px,py;                               /* Posicion no buffer de destino */
 
 if ((temporal = (bytepunt) malloc (64000)) == NULL) {
       error(non_hai_mem);
 }
 
 for (barrido=0;barrido<4;barrido++)
    for (py=inicio[barrido];py<y;py+=salto[barrido]) {
          pos_dest = x*py;
          for (px=0;px<x;px++) temporal[pos_dest++]=buffer[pos_orig++];
    }
 
 free(buffer);
 return (temporal);
} 

/***************************************************************************/

#undef ROOT
#undef CC
#undef EOI

