/*     Utilidade para ver pantallas de Spectrum e convertilas de/a GIF     */
/*                 Programada por Alberto Garcia Gonzalez                  */
/*          Modulo para traballar con archivos .PCX 256x192x256            */

/* ULTIMA REVISION: 16-9-98 18:04 */

/***************************************************************************/

bytepunt leer_pcx(char nome[],int *x, int *y)
{
 FILE *fich;                                    /* O ficheiro que imos ler */
 bytepunt buffer = NULL;                       /* A pantalla descomprimida */
 bytepunt temporal;                 /* Buffer temporal onde lemos os datos */
 dword tamanho;		                /* O tamanho dos datos comprimidos */
 byte rle;                           /* Si o PCX usa compresion RLE ou non */
 dword orix;                              /* Posicion da que estamos lendo */
 dword pos;                          /* Posicion na que estamos escribindo */
 byte reps;              /* Repeticions do byte que estamos descomprimindo */
 byte valor;                            /* Valor que estamos descomprimido */

 if ((fich = fopen(nome,"rb")) == NULL) {
       error1(err_apertura,nome);
 }

 tamanho = filelength(fileno(fich))-897; /* 128 cab + 768 paleta + 1 marca */

 if ((temporal = (bytepunt) malloc (64000)) == NULL) {
       error(non_hai_mem);
 }
 
 if (bpfread(temporal,128,1,fich) != 1) {     /* Lemos a cabeceira do .PCX */
       error(err_datos);
 }

 if ((temporal[0] != 10) || (temporal[1] != 5)) {
       error(non_e_PCX);             /* Comprobamos o identificador do PCX */
 }

 rle = temporal[2];                  /* Indica se o PCX ten compresion RLE */

 if (temporal[3] != 8) {                     /* O numero de bits por pixel */
       error(err_datos);
 }

 *x = temporal[8]  + 256 * (temporal[9]  - temporal[5]) - temporal[4] + 1;
 *y = temporal[10] + 256 * (temporal[11] - temporal[7]) - temporal[6] + 1;

 if ((*x > 256) || (*y > 192)) {
       error(tam_img_inc);         /* O tamanho maximo permitido e 256x192 */
 }

 if ((temporal[65] != 1) ||                /* A imaxe so debe ter un plano */
    (bpfread(temporal,tamanho,1,fich) != 1) ||            /* Lemos a imaxe */
    (fgetc(fich) != 12)) {   /* Comprobamos se esta a marca da paleta (12) */
       error(err_datos);     
 }
 
 leer(paleta,768,fich);               /* Lemos a paleta enteira na memoria */
 for (pos=0;pos<768;pos++) paleta[pos] >>= 2;
 fclose(fich);
 
 if (rle) {                             /* Se hai que descomprimir a imaxe */
       if ((buffer = (bytepunt) malloc(64000)) == NULL) {
             error(non_hai_mem);         /* Asignamos espacio para a imaxe */
       }
 } else return temporal;

 orix = 0;
 pos = 0;
 
 while ( pos < ((dword)(*x)*(*y)) ) {
       valor = temporal[orix++];
       if ((valor & 192) == 192) {       /* Comprobamos se hai repeticions */
             reps = (valor & 63);          /* Repeticions do seguinte byte */
             valor = temporal[orix++];
       } else reps = 1;
       for (;reps>0;reps--) {
             buffer[pos++] = valor;
       }
 }
 free(temporal);
 return buffer;
}

/***************************************************************************/

bytepunt comprimir_pcx(bytepunt orix,dword *tamanho)
{
 char cabec1[]="\xa\x5\x1\x8\0\0\0\0\xff\0\xbf\0\0\x1\xc0\0";
 char cabec2[]="\0\x1\0\x1\x1\0";
 bytepunt dest;                           /* Aqui e onde vai a estar o PCX */
 dword pos_orix;                            /* Posicion no buffer de orixe */
 dword pos_dest;                          /* Posicion no buffer de destino */
 byte valor;                               /* O valor que estamos gardando */
 byte reps;                                /* As repeticions de cada valor */
 byte restantes;                           /* Bytes restantes por scanline */
 
 if ((dest = (bytepunt) malloc (64000)) == NULL) {
       error(non_hai_mem);
 }
 
 bpmemcpy(dest,cabec1,16);
 bpmemcpy(dest+64,cabec2,6);
 for (valor=0;valor<16;valor++) {                     /* Copiamos a paleta */
       dest[16+valor*3] = colores[valor][0] << 2;
       dest[17+valor*3] = colores[valor][1] << 2;
       dest[18+valor*3] = colores[valor][2] << 2;
 }
 for (pos_dest=70;pos_dest<128;pos_dest++) dest[pos_dest]=0;
 
 /* Aqui comeza a compresion propiamente dita */

 pos_orix = 0;
 pos_dest = 128;
 restantes = 0;
 
 while (pos_orix<49152) {
       reps = 1;
       valor = orix[pos_orix++];
       if (restantes>0)
             restantes--;
             else restantes = 255;
       while ((pos_orix<49152) && (restantes>0) && (reps<63) &&
             (orix[pos_orix]==valor)) {
             reps++;
             pos_orix++;
             restantes--;
       }
       if ((reps != 1) || ((valor & 192) == 192)) {
             dest[pos_dest++] = reps | 192;
       }
       dest[pos_dest++] = valor;
 }
 
 /* Fin da compresion */
 
#ifdef _NON_O_LEE_NINGUN_PROGRAMA_
 if (pos_dest>=49280) {      /* Se ocupa mais comprimido que descomprimido */
       for (pos_dest=128;pos_dest<49280;pos_dest++) {
             dest[pos_dest]=orix[pos_dest-128];
       }
       dest[2]=0;                                /* Non hai compresion RLE */
 }
#endif
 
 dest[pos_dest++] = 12;                       /* Marca de comezo da paleta */
 for (valor=0;valor<16;valor++) {                     /* Copiamos a paleta */
       dest[pos_dest++] = colores[valor][0] << 2;
       dest[pos_dest++] = colores[valor][1] << 2;
       dest[pos_dest++] = colores[valor][2] << 2;
 }

 for (pos_orix=0;pos_orix<720;pos_orix++) dest[pos_dest++]=0;
 
 *tamanho = pos_dest;
 return dest;
}

/***************************************************************************/

