/*     Utilidade para ver pantallas de Spectrum e convertilas de/a GIF     */
/*                 Programada por Alberto Garcia Gonzalez                  */
/*                     Cabeceira para Turbo C/C++ 3.0                      */

/* ULTIMA REVISION: 3-3-99 22:38 */

#define PLATAFORMA "DOS/16"
typedef byte far * bytepunt;
typedef unsigned long dword;
typedef long sdword;
#define malloc farmalloc
#define free farfree
#include <conio.h>
#include <io.h>
#include <dos.h>

/* void (*_vaciar)(byte far *buffer); */
void (*_volcar)(byte far *buffer);
void (*_bpmemcpy)(bytepunt destino, const void *orixe, size_t n);
void vaciar8086(byte far *buffer);
void volcar8086(byte far *buffer);
void bpmemcpy8086(bytepunt destino, const void *orixe, size_t n);
void vaciar80386(byte far *buffer);
void volcar80386(byte far *buffer);
void bpmemcpy80386(bytepunt destino, const void *orixe, size_t n);
int e80386(void);

#define vaciar (*_vaciar)
#define volcar (*_volcar)
#define bpmemcpy (*_bpmemcpy)

/***************************************************************************/

void zxsee_init(void)
{
 if (e80386()) {
/*       _vaciar = &vaciar80386; */
       _volcar = &volcar80386;
       _bpmemcpy = &bpmemcpy80386;
 } else {
/*       _vaciar = &vaciar8086; */
       _volcar = &volcar8086;
       _bpmemcpy = &bpmemcpy8086;
 }
}

/***************************************************************************/

void modo13h(void)
{
 asm{
     mov ax,0x13
     int 0x10
    }
}

/***************************************************************************/

void modotexto(void)
{
 asm{
     mov ax,3
     int 0x10
    }
}

/***************************************************************************/

void putcolor (byte color, byte r, byte g, byte b)
{
 asm{
     mov dx,0x3C8
     mov al,[color]
     out dx,al
     inc dx
     mov al,[r]
     out dx,al
     mov al,[g]
     out dx,al
     mov al,[b]
     out dx,al
    }
}

/***************************************************************************/

void vaciar8086(byte far *buffer)
{
 asm{
     les di,buffer
     xor ax,ax
     mov cx,32000
     rep stosw
 }
}

/***************************************************************************/

void vaciar80386(byte far *buffer)
{
 asm{
     les di,buffer
     db 0x66; xor ax,ax
     mov cx,16000; db 0x66
     rep stosw
 }
}

/***************************************************************************/

void volcar8086(byte far *buffer)
{
 asm{
     mov bx,ds
     mov ax,0x0a000
     mov es,ax
     xor di,di
     lds si,[buffer]
     mov cx,32000
     rep movsw
     mov ds,bx
 }
}

/***************************************************************************/

void volcar80386(byte far *buffer)
{
 asm{
     mov bx,ds
     mov ax,0x0a000
     mov es,ax
     xor di,di
     lds si,[buffer]
     mov cx,16000; db 0x66
     rep movsw
     mov ds,bx
 }
}

/***************************************************************************/

void bpmemcpy8086(bytepunt destino, const void *orixe, size_t n)
{
 asm{
     les  di,[destino]
     mov  si,[orixe]
     mov  cx,n
     test cx,1
     jz   Label1
     movsb
    }
    Label1:
 asm{
     shr  cx,1
     rep movsw
    }
}

/***************************************************************************/

void bpmemcpy80386(bytepunt destino, const void *orixe, size_t n)
{
 asm{
     les  di,[destino]
     mov  si,[orixe]
     mov  cx,n
     test cx,1
     jz   Label1
     movsb
    }
    Label1:
 asm{
     shr  cx,1
     test cx,1
     jz   Label2
     movsw
    }
    Label2:
 asm{
     shr cx,1
     db 0x66; rep movsw
    }
}

/***************************************************************************/

size_t bpfread(bytepunt punt, size_t size, size_t nmemb, FILE *fich)
{
 unsigned tamanho = size*nmemb;
 unsigned tam_dos_rw;
 if ((_dos_read(fileno(fich),punt,tamanho,&tam_dos_rw)) ||
    (tamanho != tam_dos_rw)) {
       return 0;
 }
 else return nmemb;
}

/***************************************************************************/

size_t bpfwrite(bytepunt punt, size_t size, size_t nmemb, FILE *fich)
{
 unsigned tamanho = size*nmemb;
 unsigned tam_dos_rw;
 if ((_dos_write(fileno(fich),punt,tamanho,&tam_dos_rw)) ||
    (tamanho != tam_dos_rw)) {
       return 0;
 }
 else return nmemb;
}

/***************************************************************************/

void espera_vret(void)
{
 asm mov  dx,0x03DA
 Label1:
 asm{
     in   al,dx
     test al,8
     jnz  Label1
    }
 Label2:
 asm{
     in   al,dx
     test al,8
     jz   Label2
    }
}

/***************************************************************************/

int teclapulsada(void)
{
 return kbhit();
}

/***************************************************************************/

int leetecla(void)
{
 return getch();
}

/***************************************************************************/

void mostrar_pantalla(bytepunt buf1,bytepunt buf2)
{
 int i;
 while (1) {
       volcar(buf1);
       for (i=0;i<10;i++) {
             if (teclapulsada()) return;
             delay(50);
       }
       volcar(buf2);
       for (i=0;i<10;i++) {
             if (teclapulsada()) return;
             delay(50);
       }
 }
}

/***************************************************************************/

/*                        PARTE DA DETECCION DA CPU                        */

/***************************************************************************/

int valid_opcode;
void interrupt (*oldint6)(void);

/***************************************************************************/

void interrupt int6handler(void)
{
 valid_opcode = 0;
 asm{
     add word ptr ss:[bp + 18], 3
 }
}

/***************************************************************************/

int e8088(void)
{
 unsigned sp1,sp2;
 asm{
    mov [sp1], sp
    push sp
    pop [sp2]
 }
 return (sp1 != sp2);
}

/***************************************************************************/

int e80386(void)
{
 if (e8088()) return 0;
 valid_opcode = 1;
 oldint6 = getvect(6);
 setvect(6,int6handler);
 asm{
     db 0x0f
     db 0x20   /* mov edx,cr0 */
     db 0xc2
 }
 setvect(6,oldint6);
 return valid_opcode;
}

/***************************************************************************/
